# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
import pickle
import copy
import numpy as np
from exo_k.gas_mix import Gas_mix
from exo_k.atm import Atm
from exo_k.atm_2band import Atm_2band
from exo_k.util.cst import DAY, RGP
from .settings import Settings
from .tracers import Tracers
from .convection import molecular_diffusion_numba, \
                convective_acceleration_numba, moist_convective_adjustment_numba, compute_condensation_numba, \
                compute_rainout_numba
from .condensation import Condensing_species, Condensation_Thermodynamical_Parameters 

class Atm_evolution(object):
    """Model of atmospheric evolution.

    Uses exo_k.Atm class to compute radiative transfer
    """

    def __init__(self, bg_vmr=None, verbose=False, **kwargs):
        """Initializes atmospheric profiles.

        Most arguments are passed directly to exo_k.Atm class through **kwargs

        .. warning::
            Layers are counted from the top down (increasing pressure order).
            All methods follow the same convention.
        """
        self.settings = Settings()
        self.settings.set_parameters(**kwargs)
        self.header={'rad':0,'conv':1,'cond':2,'madj':3,'rain':4,'tot':5}

        # setup background gas and thermodynamical properties
        self.bg_gas = Gas_mix(bg_vmr)
        self.M_bg = self.bg_gas.molar_mass()
        self.M_bg = self.settings.use_or_set('M_bg', self.M_bg)
        self.cp = self.bg_gas.cp()
        self.cp = self.settings.pop('cp', self.cp)
        self.rcp = RGP/(self.M_bg*self.cp)
        self.rcp = self.settings.use_or_set('rcp', self.rcp)
        if (not isinstance(self.rcp, float)) or (not isinstance(self.cp, float)):
            print('None of rcp or cp should be arrays. If you provided arrays')
            print('for the background gas composition, you should also')
            print('provide the effective cp and rcp for your atmosphere.')
            raise RuntimeError('None of rcp or cp should be arrays.')
        if verbose: print('cp, M_bg, rcp:', self.cp, self.M_bg, self.rcp)

        # setup tracers
        self.tracers = Tracers(self.settings, bg_vmr = self.bg_gas.composition,
            **self.settings.parameters)
        self.initialize_condensation(**self.settings.parameters)

        self.setup_radiative_model(gas_vmr = self.tracers.gas_vmr,
            **self.settings.parameters)
        self.Nlay = self.atm.Nlay
        self.tlay = self.atm.tlay
        self.compute_hybrid_coordinates()

        if verbose: print(self.settings.parameters)
        self.evol_tau = 0.

    def set_options(self, reset_rad_model = False, check_keys = True,
            Kzz = None, cp = None, verbose = False, **kwargs):
        """This method is used to store the global options
        in the `Settings` object.

        Arguments are all passed through **kwargs.

        Sometimes, one needs to reset the radiative model to take into
        account some modifications (like the databases). Normally, this should
        be automatic, but you can force it with `reset_rad_model=True`
        """
        if check_keys:
            for key in kwargs.keys():
                if key in self.settings._forbidden_changes:
                    print('Warning!!! ', key, ' cannot be changed by set_options.')
                    print('You should probably initialize a new Atm_evolution instance.')
                    print('Use check_keys = False to remove this warning.')
        if 'tlay' not in kwargs.keys():
            self.settings.set_parameters(tlay=self.tlay, logplay=self.atm.logplay, **kwargs)
        else:
            self.settings.set_parameters(**kwargs)
        if Kzz is not None:
            self.tracers.Kzz = np.ones(self.Nlay) * Kzz
        if cp is not None:
            self.cp = cp
        if 'radiative_acceleration' in kwargs.keys():
            print("'radiative_acceleration' is deprecated. Please use acceleration_mode instead.")
            print("acceleration_mode = 1 will emulate the previous behavior but other modes exist.")
            print("In particular acceleration_mode = 4 will also accelerate convergence in convective zones")
            raise DeprecationWarning('radiative_acceleration is deprecated. Remove radiative_acceleration from the options to get rid of this message')
        
        if not set(kwargs.keys()).issubset(self.settings._non_radiative_parameters):
            reset_rad_model = True
            if verbose: print('Radiative model will be reset.')
        if reset_rad_model: self.setup_radiative_model(gas_vmr = self.tracers.gas_vmr,
                **self.settings.parameters)

    def initialize_condensation(self, condensing_species = None, **kwargs):
        """This method initializes the condensation module by
        listing all the condensing vapors and linking them to their
        condensed form. 

        For each vapor-condensate pair, a :class:`Condensible_species` object is created
        with the thermodynamical data provided. 

        Here is an example of dictionary to provide as input to include CH4 condensation
        ```
        condensing_species={'ch4':{'Latent_heat_vaporization': 5.25e5, 'cp_vap': 2.232e3, 'Mvap': 16.e-3,
            'T_ref': 90., 'Psat_ref': 0.11696e5}}
        ```
        """
        if condensing_species is None:
            condensing_species = {}
        self.condensing_pairs = list()
        self.condensing_pairs_idx = list()
        self.condensing_species_idx = dict()
        self.condensing_species_params=list()
        self.condensing_species_thermo=list()
        idx=0
        for name in self.tracers.namelist:
            if 'type' in self.tracers.dico[name]:
                if self.tracers.dico[name]['type'] == 'vapor':
                    if 'condensed_form' not in self.tracers.dico[name]:
                        print("You should identify the 'condensed_form' of:", name)
                        raise RuntimeError()
                    elif self.tracers.dico[name]['condensed_form'] not in self.tracers.namelist:
                        print("The condensed form of a vapor should be a tracer.")
                        raise RuntimeError()
                    elif name in condensing_species.keys():
                        cond_name = self.tracers.dico[name]['condensed_form']
                        self.condensing_species_idx[name]=idx
                        self.condensing_pairs.append([name, cond_name])
                        self.condensing_pairs_idx.append(\
                            [self.tracers.idx[name], self.tracers.idx[cond_name]])
                        self.condensing_species_params.append(\
                            Condensing_species(**condensing_species[name]))
                        self.condensing_species_thermo.append(\
                            Condensation_Thermodynamical_Parameters(**condensing_species[name]))
                        idx+=1
                    else: 
                        print("The thermodynamic parameters for:", name,'were not provided')
                        print('through condensing_species = {}.')
                        raise RuntimeError()
        self.Ncond=idx

    def setup_radiative_model(self, k_database=None, k_database_stellar=None,
            cia_database=None, cia_database_stellar=None, gas_vmr=None, **kwargs):
        """This method initializes the exo_k.Atm object that will be used
        to carry out radiative transfer calculations. 

        This is where the radiative data used are chosen and transmitted to the 
        radiative transfer module, along with many other
        parameters including the incoming stellar flux (`flux_top_dw`), the
        blackbody temperature of the star (`Tstar`), the 

        If a `k_database_stellar` is provided, then this is this database
        that will be used to treat the scattering and absorption of incoming radiation.
        In that case, `k_database` will be used to treat the emission of the atmosphere.
        The effective cos(zenith angle) for the incoming stellar radiation can then
        be specified independently with the `mu0_stellar` keyword.

        If no `k_database_stellar` is provided, `k_database` will be used to treat
        both the atmospheric emission and the stellar radiation. Running a model
        with `k_database_stellar=k_database` yields the same results at twice the cost.
        
        Parameters
        ----------
            k_database, k_database_stellar: `exo_k.Kdatabase` objects
                radiative database for the molecules in the atmospheres.
            cia_database, cia_database_stellar: `exo_k.CIA_database` object
                radiative database for the CIA ofmolecules in the atmospheres.
        """
        if k_database is None:
            raise RuntimeError('We need at least a k_database')
        if k_database_stellar is None:
            self.atm = Atm(k_database=k_database, cia_database=cia_database, composition=gas_vmr, **kwargs)
        else:
            raise DeprecationWarning("k_database_stellar is deprecated. Proceed at your own risk.")
            self.atm = Atm_2band(k_database=k_database, cia_database=cia_database,
                k_database_stellar=k_database_stellar, cia_database_stellar=cia_database_stellar,
                composition=gas_vmr, **kwargs)
        H, net = self.atm.heating_rate(compute_kernel=True, **kwargs)

    def compute_average_fluxes(self):
        """Use the averaged heating rates to compute the various fluxes (W/m^2)
        at the level interfaces. These fluxes are positive when the energy flows
        upward.

        To be consistent with radiative fluxes, the first array value corresponds
        to the top of atmosphere and should be 0 in most cases. The last value corresponds
        to the flux between the deepest layer (considered to be the surface) and the layer just above.
        """
        self.Fnet = np.zeros((6, self.Nlay))
        self.Fnet[0] = self.Fnet_rad
        for ii in range(1,5):
            self.Fnet[ii]=np.concatenate([[0.],
                np.cumsum(self.H_ave[ii]*self.atm.dmass)[:-1]])
        self.Fnet[-1] = np.sum(self.Fnet, axis=0)


    def evolve(self, N_timestep=1, N_kernel=10000, timestep_factor=1., dT_max = 100., verbose = False, **kwargs):
        r"""The time variable used in the model is tau=t/cp. 
        The equation we are solving in each layer is thus

        .. math::
            c_p \frac{d T}{d t}= \frac{d T}{d tau} = \sum_i H_i

        For a given timestep `dtau`, the physical time elapsed in second can be computed using `dt=dtau*cp`

        To work, the heating rates (H) must be computed in W/kg. 

        This also means that if one needs the physical rate of change of another quantity (like dq/dt)
        from the delta q over a time step,
        one needs to do `dq/dt = delta q / (timestep * cp)`

        Parameters
        ----------
            N_timestep: int
                Number of timesteps to perform.
            N_kernel: int
                Maximal number of timesteps between two computations of the radiative kernel.
            timestep_factor: float
                Multiplicative factor applied to timestep computed automatically by
                the radiative module.
                timestep_factor > 1 can lead to unstabilities.
            dT_max: float
                Maximum temperature increment in a single timestep.
            
        """
        if self.atm.k_database is None:
            print('This Atm_evolution instance is not linked to any k_database')
            print('Use self.set_options(k_database=, ...)')
            raise RuntimeError('No k_database provided.')
        self.H_ave = np.zeros((6, self.Nlay))

        self.tlay_hist = np.zeros((N_timestep,self.Nlay))
        self.Fnet_top = np.zeros((N_timestep))
        self.timestep_hist = np.zeros((N_timestep))
        tau0 = self.evol_tau
        self.N_last_ker = 0
        compute_kernel = False
        dTlay_max = 2. * self.settings['dTmax_use_kernel']
        self.tracers.update_gas_composition(update_vmr=True)
        for ii in range(N_timestep):
            if np.amax(np.abs(self.tlay-self.atm.tlay_kernel)) < self.settings['dTmax_use_kernel']:
                self.N_last_ker +=1
                if verbose: print(self.N_last_ker, self.N_last_ker%N_kernel)
                compute_kernel = (self.N_last_ker%N_kernel == 0)
            else:
                if dTlay_max < 0.5 * self.settings['dTmax_use_kernel']:
                    compute_kernel = True
                    self.N_last_ker = 0
                else:
                    compute_kernel = False
                    self.N_last_ker +=1
            if ii == N_timestep-1:
                if ii != 0:
                    compute_kernel=True
            self.H_tot=np.zeros(self.Nlay)
            if verbose: print('iter, compute_kernel:', ii, compute_kernel)
            if self.tracers.some_var_gases:
                gas_vmr_rad = self.tracers.gas_vmr
            else:
                gas_vmr_rad = None
            self.H_rad, self.Fnet_rad = self.atm.heating_rate(compute_kernel = compute_kernel,
                rayleigh = self.settings['rayleigh'], dTmax_use_kernel=self.settings['dTmax_use_kernel'],
                gas_vmr = gas_vmr_rad, **kwargs)
#            if verbose and compute_kernel: print('H_rad', self.H_rad)
            self.H_tot += self.H_rad
            self.timestep = timestep_factor * self.atm.tau_rad
            #if verbose: print('tau_rad, dt:', self.atm.tau_rad, self.timestep)
            self.evol_tau += self.timestep
            if self.settings['convection']:
                self.H_conv = self.tracers.dry_convective_adjustment(self.timestep, self.H_tot, self.atm, verbose = verbose)
                self.H_tot += self.H_conv
            else:
                self.H_conv = np.zeros(self.Nlay)
            if self.settings['diffusion']:
                self.tracers.turbulent_diffusion(self.timestep, self.H_tot, self.atm, self.cp)
                self.tracers.update_gas_composition(update_vmr=False)
            if self.settings['molecular_diffusion']:
                self.H_diff = self.molecular_diffusion(self.timestep,
                    self.H_tot, self.atm, self.cp)
                self.H_tot += self.H_diff
            qarray_before_condensation = np.copy(self.tracers.qarray)
            if self.settings['moist_convection']:
                self.H_madj = self.moist_convective_adjustment(self.timestep, self.H_tot,
                                    moist_inhibition=self.settings['moist_inhibition'], verbose = verbose)                
                self.H_tot += self.H_madj
            else:
                self.H_madj = np.zeros(self.Nlay)
            if self.settings['condensation']:
                self.H_cond = self.condensation(self.timestep, self.H_tot, verbose = verbose)
                self.H_tot += self.H_cond
            else:
                self.H_cond = np.zeros(self.Nlay)
            if self.settings['rain']:
                self.H_rain = self.rainout(self.timestep, self.H_tot, verbose = verbose)
                self.H_tot += self.H_rain
            else:
                self.H_rain = np.zeros(self.Nlay)
            if self.settings['mass_redistribution']:
                self.mass_redistribution(qarray_before_condensation)
            if self.settings['surface_reservoir']:
                self.tracers.update_surface_reservoir(condensing_pairs_idx = self.condensing_pairs_idx,
                    surf_layer_mass = self.atm.dmass[-1])
            if self.settings['acceleration_mode'] > 0:
                self.radiative_acceleration(timestep = self.timestep, \
                    acceleration_mode = self.settings['acceleration_mode'], verbose = verbose)
                #self.H_tot *= self.acceleration_factor
            dTlay= self.H_tot * self.timestep
            dTlay_max = np.amax(np.abs(dTlay))
            if dTlay_max > dT_max:
                print('dT > dTmax:',dTlay_max,' at k=',np.argmax(np.abs(dTlay))) 
            if verbose:
                print('heat rates (rad, dry conv), dTmax:', 
                    np.sum(self.H_rad*self.atm.dmass), np.sum(self.H_conv*self.atm.dmass),
                    dTlay_max)
            dTlay=np.clip(dTlay,-dT_max,dT_max)
            self.tlay = self.tlay + dTlay
            self.tlay_hist[ii] = self.tlay
            for jj, H in enumerate([self.H_rad, self.H_conv, self.H_cond, self.H_madj,
                  self.H_rain, self.H_tot]):
                self.H_ave[jj] += H * self.timestep
            self.Fnet_top[ii] = self.Fnet_rad[0]
            self.timestep_hist[ii] = self.timestep
            self.atm.set_T_profile(self.tlay)
            self.tracers.update_gas_composition(update_vmr=True)
        inv_delta_t = 1./(self.evol_tau-tau0)
        self.H_ave *= inv_delta_t
        self.compute_average_fluxes()

    def equilibrate(self, Fnet_tolerance = None, N_iter_max = 10,
        N_timestep_ini = 100, N_timestep_max = 1000000, verbose = False, **kwargs):
        """Evolves an atmosphere until it is at equilibrium.
        
        Equilibrium is assumed to be reached when the net top of atmosphere
        flux remains within +-Fnet_tolerance of the internal flux
        for a whole evolution step.

        The number of timesteps per evolution step in multiplied by two 
        at each iteration, starting from N_timestep_ini, until the limit of
        N_timestep_max is reached.

        Parameters
        ----------
            Fnet_tolerance: float
                Tolerance on net flux in W/m^2 to identify convergence.
            N_iter_max: int
                Max number of successive calls to evolve
            N_timestep_ini: int
                Initial number of timesteps in a single evolution step
            N_timestep_max: int
                Max number of timesteps in a single evolution step
        """
        iter=1
        if Fnet_tolerance is None:
            raise RuntimeError('You should provide the maximum tolerance on the net flux: Fnet_tolerance (in W/m^2)') 
        N_timestep = N_timestep_ini
        while iter <= N_iter_max:
            time_init = self.evol_tau
            self.evolve(N_timestep = N_timestep, **kwargs)
            net = self.Fnet_top - self.atm.internal_flux
            if verbose:
                print('iter: {iter}, N_timestep: {Nt}'.format(iter = iter, Nt = N_timestep))
                print('Fnet mean: {fme:.3g} W/m^2, (min:{fmi:.3g}, max:{fma:.3g})'.format( \
                    fme = net.mean(), fmi = net.min(), fma = net.max()))   
                print('timestep: {ts1:.3g} d | {ts2:.3g} s, total time: {ev_t:.3g} yr'.format( \
                    ts1 = self.timestep*self.cp/(DAY), ts2 = self.timestep*self.cp,
                    ev_t = (self.evol_tau-time_init)*self.cp/(DAY*365.)))   
                #print('timestep:',self.timestep*self.cp/(DAY),'days, ',
                #    self.timestep*self.cp,'s, evol_time(yr):',(self.evol_tau-time_init)*self.cp/(DAY*365.))
            if np.all(np.abs(net) < Fnet_tolerance): break
            N_timestep = min( N_timestep * 2, N_timestep_max)
            iter += 1
            
    def moist_convective_adjustment(self, timestep, Htot, moist_inhibition = True,
            verbose = False):
        """This method computes the vapor and temperature tendencies do to
        moist convectoin in saturated layers.

        The tracer array in modified in place.

        Parameters
        ----------
            timestep: float
                physical timestep of the current step (in s/cp).
            Htot: array
                Total heating rate (in W/kg) of all physical processes
                already computed

        Return
        ------
            H_madj: array
                Heating rate due to large scale condensation (W/kg)
        """
        new_t = self.atm.tlay + timestep * Htot
        H_madj = np.zeros(self.Nlay)
        for i_cond in range(self.Ncond): #careful i_cond is the index of the condensing pair
            # in the list of condensing species, idx_cond is the position of the
            # condensate linked to i_cond in the tracers array.
            idx_vap, idx_cond = self.condensing_pairs_idx[i_cond]
            thermo_parameters = self.condensing_species_thermo[i_cond].th_params
            H, qarray, new_t = moist_convective_adjustment_numba(timestep, self.Nlay,
                new_t, self.atm.play, self.atm.dmass, self.cp, self.tracers.Mgas, self.tracers.qarray, idx_vap, idx_cond,
                thermo_parameters,
                moist_inhibition = moist_inhibition, verbose = verbose)
            #print('t after madj:', new_t)
            H_madj += H
            self.tracers.qarray = qarray
            if verbose: print(qarray[idx_cond])
        return H_madj

    def molecular_diffusion(self, timestep, Htot, atm, cp):
        """Mixes energy following a diffusion equation
        with a constant Dmol parameter (self.Dmol in m^2/s).

        Parameters
        ----------
            timestep: float
                physical timestep of the current step (in s/cp).
                (needs to be converted before it is sent to `turbulent diffusion)
            Htot: array
                Total heating rate (in W/kg) of all physical processes
                already computed
            atm: :class:`Atm` object
                The Atm object used in the radiative transfer which
                contains many state variables. 
        """
        new_t = atm.tlay + timestep * Htot
        H_diff = molecular_diffusion_numba(timestep*cp, self.Nlay,
                    atm.play, atm.plev,
                    atm.dmass, new_t, self.tracers.Mgas,
                    atm.grav, self.tracers.Dmol)
        return H_diff

    def condensation(self, timestep, Htot, verbose = False):
        """This method computes the vapor and temperature tendencies do to
        large scale condensation in saturated layers.

        The tracer array in modified in place.

        Parameters
        ----------
            timestep: float
                physical timestep of the current step (in s/cp).
            Htot: array
                Total heating rate (in W/kg) of all physical processes
                already computed

        Return
        ------
            H_cond: array
                Heating rate due to large scale condensation (W/kg)
        """
        new_t = self.atm.tlay + timestep * Htot
        H_cond = np.zeros(self.Nlay)
        for i_cond in range(self.Ncond): #careful i_cond is a dumy loop index, idx_cond is position of species i_cond in tracers array.
            idx_vap, idx_cond = self.condensing_pairs_idx[i_cond]
            thermo_parameters = self.condensing_species_thermo[i_cond].th_params
            H_cond += compute_condensation_numba(timestep, self.Nlay, new_t, self.atm.play,
                self.cp, self.tracers.Mgas, self.tracers.qarray,
                idx_vap, idx_cond, thermo_parameters,
                latent_heating = self.settings['latent_heating'],
                condensation_timestep_reducer = self.settings['condensation_timestep_reducer'],
                verbose = verbose)
        return H_cond

    def rainout(self, timestep, Htot, verbose = False):
        """This method computes rainout.

        Condensates are carried down and reevaporated whenever there is
        "room" in an unsaturated layer. 

        The option `evap_coeff` acts has an efficiency factor. `evap_coeff=1`
        is the efficient evaporation limit. When `evap_coeff<1` the maximum amount of
        condensates that can be reevaporated in a single layer is multiplied by
        `evap_coeff`

        All condensates are finaly evaporated in the last layer or when T > Tboil.

        The tracer array is modified in place.

        Parameters
        ----------
            timestep: float
                physical timestep of the current step (in s/cp).
            Htot: array
                Total heating rate (in W/kg) of all physical processes
                already computed

        Return
        ------
            H_rain: array
                Heating rate due to re evaporation (W/kg)
        """
        new_t = self.atm.tlay + timestep * Htot
        H_rain=np.zeros(self.Nlay)
        for i_cond in range(self.Ncond):
            idx_vap, idx_cond = self.condensing_pairs_idx[i_cond]
            thermo_parameters = self.condensing_species_thermo[i_cond].th_params
            H_rain += compute_rainout_numba(timestep, self.Nlay, new_t, self.atm.play,
                self.atm.dmass, self.cp, self.tracers.Mgas, self.tracers.qarray,
                idx_vap, idx_cond, thermo_parameters,
                self.settings['evap_coeff'], self.tracers.qdeep[idx_vap], verbose = verbose)

        return H_rain

    def compute_hybrid_coordinates(self):
        """Compute hybrid coordinates as in GCM.

        This will be used when surface pressure changes.
        
        Convention : sigma = (p-ptop)/(psurf-ptop)
        For each layer/level, the pressure is p = sigma * psurf + gamma
        """
        psurf = self.atm.plev[-1]
        ptop = self.atm.plev[0]
        self.sigma_lay = (self.atm.play-ptop)/(psurf-ptop)
        self.gamma_lay = (1.-self.sigma_lay)*ptop
        self.sigma_lev = (self.atm.plev-ptop)/(psurf-ptop)
        self.gamma_lev = (1.-self.sigma_lev)*ptop
        self.dsigma_lev = np.diff(self.sigma_lev)

    def compute_mass_flux(self, dvapor_mass, sum_dvapor_mass):
        """Computes the mass flux through the hybrid coordinate interfaces (kg/s/m^2; positive upward).
        (see Methods in Leconte et al. (2013))

        W[k] is the mass flux between layer k-1 et k.

        Parameters
        ----------
            sum_dvapor_mass: float
                Total mass of vapor added to the atmosphere
                in the last timestep.
            dvapor_mass: array
                mass of vapor added to each layer.

        For the moment, W[0] = W[Nlay]
        """
        W = np.zeros(self.Nlay+1)
        #W[0] = 0. #No mass flux between the top of the atm and space
        W[1:] = np.cumsum(self.dsigma_lev * sum_dvapor_mass - dvapor_mass) 
        W[-1] = 0. #No mass flux between the surface and the atm
        return W

    def mass_redistribution(self, qarray_before_condensation):
        """Update new mass and new pressure of a layer due to the evaporation and condensation of a given species, 
        for more details see Methods, Leconte et al., 2013 (Nature)
     
        Parameters
        ----------

        """
        if self.Ncond > 1:
            raise NotImplementedError('Mass redistribution limited to one condensing species')
        #Pour juste l'eau 
        for i_cond in range(self.Ncond):
            idx_vap, idx_cond = self.condensing_pairs_idx[i_cond]
 
        dq_mass_redist = np.zeros_like(qarray_before_condensation)

        variation_qarray = self.tracers.qarray - qarray_before_condensation #Evolution of qarray before and after condensation/rain/moist_convection steps
        
        dvapor_mass = self.atm.dmass*variation_qarray[idx_vap] #Need to use idx_vap because dvapor_mass
        sum_dvapor_mass = np.sum(dvapor_mass)
        dcond_mass = self.atm.dmass*variation_qarray[idx_cond]
        self.dpsurf = sum_dvapor_mass*self.atm.grav
        dgas_mass = self.dsigma_lev*self.dpsurf/self.atm.grav
            
        if self.settings['compute_mass_fluxes']:
            self.W = self.compute_mass_flux(dvapor_mass, sum_dvapor_mass)
            for indice, q in enumerate(self.tracers.qarray):
                if indice == idx_vap: #Condensible gas in vapor form
                    epsilon = dvapor_mass
                elif indice == idx_cond: #Condensible gas in condensed form
                    epsilon = dcond_mass 
                else: #Other tracers
                    epsilon = 0.
                qarray_lev = np.zeros(self.Nlay+1)
                qarray_lev[1:-1] = (q[1:] + q[:-1]) / 2 #We choose the arithmetic mean value of the qarray as the value of the qarray at the middle of the levels ie q_(k+1/2)
                qarray_transport_through_sigma_lev = (qarray_lev[1:]-qarray_before_condensation[indice])*self.W[1:] \
                    - (qarray_lev[:-1]-qarray_before_condensation[indice])*self.W[:-1] #Attention si W à la surface est non nulnp.diff(self.qarray_lev*self.W)
                dq_mass_redist[indice] = (1./(self.atm.dmass+dgas_mass))* \
                    (qarray_transport_through_sigma_lev+epsilon-qarray_before_condensation[indice]*dvapor_mass)
                self.tracers.qarray[indice] = np.abs(qarray_before_condensation[indice] + dq_mass_redist[indice]) # carefull abs should be removed
        else:
            self.W = np.zeros(self.Nlay+1)
            for indice, q in enumerate(qarray_before_condensation):
                if indice==idx_vap: #Condensible gas in vapor form
                    epsilon = dvapor_mass
                elif indice==idx_cond: #Condensible gas in condensed form
                    epsilon = dcond_mass
                else: #Other tracer
                    epsilon = 0.
                dq_mass_redist[indice] = (1./(self.atm.dmass+dgas_mass))*(epsilon-q*dgas_mass)
                self.tracers.qarray[indice] = qarray_before_condensation[indice] + dq_mass_redist[indice]
                
        plev = self.sigma_lev*(self.atm.psurf+self.dpsurf)+self.gamma_lev
        play = self.sigma_lay*(self.atm.psurf+self.dpsurf)+self.gamma_lay
        self.atm.update_pressure_profile(play = play, plev = plev)

    def radiative_acceleration(self, timestep = 0., acceleration_mode = 0, verbose = False, **kwargs):
        """"Computes an acceleration factor and a new heating rate to speed up convergence

        Parameters
        ----------
            acceleration_mode: int
                0: no acceleration
                1 or 3: acceleration limited to radiative zones.
                2 or 4: acceleration in convective zones as well. 

        * 1: acceleration limited to radiative zones. 
          The largest radiative timescale in non-radiative zones is 
          used as reference radiative timsescale to compute acceleration.
        * 2: Same as mode 1 + acceleration in convective zones
        * 3 acceleration limited to radiative zones.
          The smallest radiative timescale in radiative zones is 
          used as reference radiative timsescale to compute acceleration.
        * 4: Same as mode 3 + acceleration in convective zones
        """
        self.acceleration_factor = np.ones_like(self.H_tot)
        rad_layers =  np.isclose(self.H_tot, self.H_rad, atol=0.e0, rtol=1.e-10)
        # determines which layer is purely radiative
        if verbose: 
            print('in acc rad_layers, H_tot, H_rad')
            print(rad_layers)
            print(np.transpose([rad_layers, self.H_tot,self.H_rad]))
        if (np.all(rad_layers)) or (not(np.any(rad_layers))):
            self.base_timescale = self.atm.tau_rad
            # we do not use timestep to avoid including the timestep_factor
        else:
            if acceleration_mode <= 2:
                self.base_timescale = np.amax(self.atm.tau_rads[np.logical_not(rad_layers)])
            elif acceleration_mode <= 4:
                self.base_timescale = np.amin(self.atm.tau_rads[rad_layers])
            else:
                raise NotImplementedError('Only acceleration_mode <= 4 is supported for the moment')

        self.acceleration_factor[rad_layers] = np.core.umath.maximum(
            self.atm.tau_rads[rad_layers] \
            / self.base_timescale * self.settings['radiative_acceleration_reducer'],
            1.)

        if (acceleration_mode == 2) or (acceleration_mode == 4):
            self.H_acc = convective_acceleration_numba(timestep, self.Nlay, self.H_rad,
                    rad_layers, self.atm.tau_rad, self.atm.tau_rads, self.atm.dmass,
                    convective_acceleration_mode = self.settings['convective_acceleration_mode'],
                    verbose = verbose)
        else:
            self.H_acc = np.zeros(self.Nlay)
        
        if verbose: 
            print('in acc acceleration_factor, H_tot, H_acc')
            print(np.transpose([self.acceleration_factor, self.H_tot, self.H_acc]))
        self.H_tot = self.H_tot * self.acceleration_factor + self.H_acc

    @property
    def time(self):
        """Yields current time in seconds
        """
        return self.evol_tau * self.cp

    @property
    def time_hist(self):
        """Yields the array of the times for the last call to evolve (in seconds)
        """
        return np.cumsum(self.timestep_hist) * self.cp
    
    def heating_rate(self, physical_process):
        """Returns heating rates in W/kg per layer averaged over last call to evolve.
        Possible physical_processes are rad, cond, conv, rain, madj, tot"""
        return self.H_ave[self.header[physical_process]]

    def net_flux(self, physical_process):
        """Returns net_flux in W/m^2 averaged over last call to evolve.
        Possible physical_processes are rad, cond, conv, rain, madj, tot"""
        return self.Fnet[self.header[physical_process]]
    
    def qsat(self, mol):
        """Returns the saturation specific concentration of molecule mol (kg/kg)"""
        cond_species_param = self.condensing_species_params[self.condensing_species_idx[mol]]
        psat = cond_species_param.Psat(self.atm.tlay)
        qsat = cond_species_param.qsat(psat, self.atm.play,
                            cond_species_param.Mvap/self.tracers.Mgas)
        return qsat

    def write_pickle(self, filename, data_reduction_level = 1):
        """Saves the instance in a pickle file

        Parameters
        ----------
            filename: str
                Path to pickle file
            data_reduction_level: int
                Level of data to delete.
                0: keep everything (results in big files).
                1: removes some arrays, should not affect subsequent evolution.
                2: removes the k and cia databases. The radiative model will need to be reset.
                This can be done with
                `set_options(k_database=, cia_database=, reset_rad_model=True)` after
                re-loading the `Atm_evolution` instance.
        """
        other = copy.deepcopy(self)
        if data_reduction_level >=1 :
            other.tlay_hist = None
            other.atm.asym_param = None
            other.atm.kdata = None
            other.atm.tau = None
            other.atm.dtau = None
            other.atm.flux_down_nu = None
            other.atm.flux_net_nu = None
            other.atm.flux_up_nu = None
            other.atm.piBatm = None
            other.atm.single_scat_albedo = None
            other.atm.gas_mix.kdata_scat = None
        if data_reduction_level >=2 :
            other.settings['k_database'] = None
            other.settings['cia_database'] = None
            other.atm.k_database = None
            other.atm.gas_mix.k_database = None
            other.atm.gas_mix.cia_database = None
            other.atm.kernel = None
            other.atm.tlay_kernel = None
            other.atm.H_kernel = None
        with open(filename, 'wb') as filehandler:
            pickle.dump(other, filehandler)



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

This module contain classes to handle atmospheric profiles.
Radiative properties are handled in atm.py which contains a daughter class.

The nomenclature for layers, levels, etc., can be found in atm.py.
"""
import numpy as np
import warnings
import astropy.units as u
from numba.typed import List
from .gas_mix import Gas_mix
from .aerosols import Aerosols
from .util.cst import N_A, PI, RGP, KBOLTZ


class Atm_profile(object):
    """A class defining an atmospheric PT profile with some global data
    (gravity, etc.)

    The derived class :class:`~exo_k.atm.Atm` handles
    radiative transfer calculations.

    """
    
    def __init__(self, composition=None, psurf=None, ptop=None, logplay=None, tlay=None,
            Tsurf=None, Tstrat=None, grav=None, Rp=None, Mgas=None, rcp=None, Nlay=20,
            logplev=None, aerosols=None,
            ## old parameters that should be None. THey are left here to catch
            ## exceptions and warn the user that their use is obsolete
            Nlev=None, tlev=None,
            **kwargs):
        """Initializes atmospheric profiles

        Parameters
        ----------
            composition: dict
                Keys are molecule names and values the vmr.
                Vmr can be arrays of size Nlev-1 (i.e. the number of layers).
            grav: float
                Planet surface gravity (gravity constant with altitude for now).
            Rp: float or Astropy.unit quantity
                Planet radius. If float, meters are assumed.
            rcp: float
                Adiabatic lapse rate for the gas (R/cp)
            Mgas: float, optional
                Molar mass of the gas (kg/mol). If given, overrides the molar mass computed
                from composition.
        
        There are two ways to define the profile.
        You can define:

        * Nlay: int
          Number of layers
        * psurf, Tsurf: float
          Surface pressure (Pa) and temperature 
        * ptop: float
          Pressure at the top of the model (Pa) 
        * Tstrat: float
          Stratospheric temperature        

        This way you will have an adiabatic atmosphere with Tsurf at the ground that
        becomes isothermal wherever T=<Tstrat.
        You can also specify:

        * logplay or play: array
        * tlay: array (same size)
          These will become the pressures (Pa; the log10 if you give
          logplay) and temperatures of the layers.
          This will be used to define the surface and top pressures.
          Nlay becomes the size of the arrays. 

        .. warning::
            Layers are counted from the top down (increasing pressure order).
            All methods follow the same convention.
        """
        if (Nlev is not None) or (tlev is not None):
            print("""
                since version 1.1.0, Nlev, tlev, and plev have been renamed
                Nlay, tlay, and logplay for consistency with other codes.
                Just change the name of the variables in the method call
                and you should be just fine!
                """)
            raise RuntimeError('Unknown keyword argument in __init__')
        self.gas_mix = None
        if composition is None: composition = dict()
        self.set_gas(composition, compute_Mgas=False)
        self.aerosols = None
        self.set_aerosols(aerosols)
        self.set_rcp(rcp=rcp)
        self.logplev = None
        self.grav = None
        if logplay is None:
            self.Nlay = Nlay
            self.Nlev = Nlay+1
            self.logplay = np.linspace(np.log10(ptop),np.log10(psurf),num=self.Nlay)
            self.compute_pressure_levels()
            self.set_adiab_profile(Tsurf=Tsurf, Tstrat=Tstrat)
        else:
            self.set_logPT_profile(logplay, tlay, logplev=logplev)
        self.set_Rp(Rp)        
        self.set_grav(grav)
        self.set_Mgas(Mgas=Mgas)

    def set_logPT_profile(self, logplay, tlay, logplev=None):
        """Set the logP-T profile of the atmosphere with a new one

        Parameters
        ----------
            logplay: array
                Log pressure (in Pa) of the layer
            tlay: array (same size)
                temperature of the layers.
        
        Other Parameters
        ----------------
            logplev: array (size Nlay+1)
                If provided, allows the user to choose the location
                of the level surfaces separating the layers.
        """
        self.logplay=np.array(logplay, dtype=float)
        self.Nlay=self.logplay.size
        self.Nlev=self.Nlay+1
        if logplev is not None:
            if logplev.size == self.Nlev:
                self.logplev=np.array(logplev, dtype=float)
            else:
                raise RuntimeError('logplev does not have the size Nlay+1')
        self.compute_pressure_levels()
        self.set_T_profile(tlay)

    def set_T_profile(self, tlay):
        """Reset the temperature profile without changing the pressure levels
        """
        tlay=np.array(tlay, dtype=float)
        if tlay.shape != self.logplay.shape:
            raise RuntimeError('tlay and logplay should have the same size.')
        self.tlay=tlay
        self.t_opac=(self.tlay[:-1]+self.tlay[1:])*0.5
        self.gas_mix.set_logPT(logp_array=self.logp_opac, t_array=self.t_opac)

    def compute_pressure_levels(self):
        """Computes various pressure related quantities
        """
        if self.logplay[0] >= self.logplay[-1]:
            print("""
            Atmospheres are modelled from the top down.
            All arrays should be ordered accordingly
            (first values correspond to top of atmosphere)""")
            raise RuntimeError('Pressure grid is in decreasing order!')
        self.play=np.power(10., self.logplay)
        if self.logplev is None:
        # case where the levels are halfway between layer centers
            self.plev=np.zeros(self.Nlev)
            self.plev[1:-1]=(self.play[:-1]+self.play[1:])*0.5
            # we choose to use mid point so that there is equal mass in the bottom half
            # of any top layer and the top half of the layer below. 
            self.plev[0]=self.play[0]
            self.plev[-1]=self.play[-1]
            ## WARNING: Top and bottom pressure levels are set equal to the
            #  pressure in the top and bottom layers. If you change that,
            #  some assumptions here and there in the code may break down!!!
            self.logplev=np.log10(self.plev)

        # case where the levels are halfway between layer centers in LOG10
            #self.logplev=np.zeros(self.Nlev)
            #self.logplev[1:-1]=(self.logplay[:-1]+self.logplay[1:])*0.5
            #self.logplev[0]=self.logplay[0]
            #self.logplev[-1]=self.logplay[-1]
            #self.plev=np.power(10.,self.logplev)
        else:
            self.plev=np.power(10., self.logplev)

        self.logp_opac=self.logplev[1:-1]
        self.psurf=self.plev[-1]
        self.dp_lay=np.diff(self.plev) ### probably redundant with dmass
        self.exner=(self.play/self.psurf)**self.rcp
        self.compute_layer_masses()

    def update_pressure_profile(self, play = None, plev = None):
        """Updates pressure levels without changing temperatures.
        
        To be used Atm_evolution class.
        """
        play = np.array(play, dtype=float)
        if play.size != self.Nlay:
            raise RuntimeError("You cannot change the number of layers in update_pressure_profile")
        self.play = play
        self.plev = np.array(plev, dtype=float)
        self.logplay = np.log10(self.play)
        self.logplev = np.log10(self.plev)
        self.logp_opac = self.logplev[1:-1]
        self.psurf = self.plev[-1]
        self.dp_lay = np.diff(self.plev) ### probably redundant with dmass
        self.exner = (self.play/self.psurf)**self.rcp
        self.compute_layer_masses()

    def set_adiab_profile(self, Tsurf=None, Tstrat=None):
        """Initializes the logP-T atmospheric profile with an adiabat with index R/cp=rcp

        Parameters
        ----------
            Tsurf: float
                Surface temperature.
            Tstrat: float, optional
                Temperature of the stratosphere. If None is given,
                an isothermal atmosphere with T=Tsurf is returned.
        """
        if Tstrat is None: Tstrat=Tsurf
        self.tlay=Tsurf*self.exner
        self.tlay=np.where(self.tlay<Tstrat,Tstrat,self.tlay)
        self.t_opac=(self.tlay[:-1]+self.tlay[1:])*0.5
        self.gas_mix.set_logPT(logp_array=self.logp_opac, t_array=self.t_opac)

    def set_grav(self, grav=None):
        """Sets the surface gravity of the planet

        Parameters
        ----------
            grav: float
                surface gravity (m/s^2)
        """
        if grav is None: raise RuntimeError('A planet needs a gravity!')
        self.grav=grav
        self.compute_layer_masses()

    def compute_layer_masses(self):
        """compute_layer_masses
        """
        if self.grav is not None:
            self.dmass=self.dp_lay/self.grav
            self.inv_dmass=1./self.dmass
    
    def set_gas(self, composition_dict, Mgas=None, compute_Mgas=True):
        """Sets the composition of the atmosphere.

        The composition_dict gives the composition in the layers, but we will
        need the composition in the radiative layers, so the interpolation is
        done here. For the moment we do a geometrical average.

        .. important::
            For the first initialization, compute_Mgas must be False
            because we need Gas_mix to be initialized before we know the number of layers
            in the atmosphere, but the number of layers is needed by set_Mgas!
            So set_Mgas needs to be called at the end of the initialization.

        Parameters
        ----------
            composition_dict: dictionary
                Keys are molecule names, and values are volume mixing ratios.
                A 'background' value means that the gas will be used to fill up to vmr=1
                If they do not add up to 1 and there is no background gas_mix,
                the rest of the gas_mix is considered transparent.
            compute_Mgas: bool
                If False, the molar mass of the gas is not updated. 
        """
        vmr_midlevel_dict=dict()
        for mol, vmr in composition_dict.items():
            if isinstance(vmr,(np.ndarray, list)):
                tmp_vmr=np.array(vmr)
                # geometrical average:
                with np.errstate(invalid='raise'):
                    try:
                        vmr_midlevel_dict[mol] = np.sqrt(tmp_vmr[1:]*tmp_vmr[:-1])
                    except FloatingPointError:
                        print('inset_gas, mol, tmp_vmr=',mol,tmp_vmr)
                        vmr_midlevel_dict[mol] = tmp_vmr[1:]
            else:
                vmr_midlevel_dict[mol] = vmr
        if self.gas_mix is None:
            self.gas_mix = Gas_mix(vmr_midlevel_dict)
        else:
            self.gas_mix.set_composition(vmr_midlevel_dict)
        if compute_Mgas: self.set_Mgas(Mgas=Mgas)

    def set_Mgas(self, Mgas=None):
        """Sets the mean molar mass of the atmosphere.

        Parameters
        ----------
            Mgas_rad: float or array of size Nlay-1
                Mean molar mass in the radiative layers (kg/mol).
                If None is given, the mmm is computed from the composition.
        """
        if Mgas is not None:
            self.Mgas_rad=Mgas
        else:
            self.Mgas_rad=self.gas_mix.molar_mass()
        if not isinstance(self.Mgas_rad, np.ndarray):
            self.Mgas_rad=self.Mgas_rad*np.ones(self.Nlay-1, dtype=float)

    def set_rcp(self, rcp = None):
        """Sets the adiabatic index of the atmosphere

        Parameters
        ----------
            rcp: float
                R/c_p
        """
        if rcp is None:
            raise RuntimeError('rcp should not be None.')
        elif not isinstance(rcp, float):
            raise RuntimeError('rcp should be a float (not an array).')
        else:
            self.rcp=rcp

    def set_aerosols(self, aerosols):
        """Sets the aerosols dictionary

        performs the interlayer averaging so that we only have properties at
        the middle of radiative layers
        """
        if aerosols is None:
            aerosols = dict()
        for aer, [reff, densities] in aerosols.items():
            if isinstance(reff,(np.ndarray, list)):
                tmp_reff=np.array(reff)
                aerosols[aer][0]=0.5*(tmp_reff[1:]+tmp_reff[:-1])
            if isinstance(densities,(np.ndarray, list)):
                tmp_densities=np.array(densities)
                #geometrical average:
                aerosols[aer][1]=np.sqrt(tmp_densities[1:]*tmp_densities[:-1])
        if self.aerosols is None:
            self.aerosols = Aerosols(aerosols)
        else:
            self.aerosols.set_aer_reffs_densities(aer_reffs_densities=aerosols)


    def set_Rp(self, Rp):
        """Sets the radius of the planet

        Parameters
        ----------
            Rp: float
                radius of the planet (m)
        """
        if Rp is None:
            self.Rp = None
            return
        if isinstance(Rp,u.quantity.Quantity):
            self.Rp=Rp.to(u.m).value
        else:
            self.Rp=Rp

    def set_Rstar(self, Rstar):
        """Sets the radius of the star

        Parameters
        ----------
            Rstar: float
                radius of the star (m)
        """
        if Rstar is None:
            self.Rstar = None
            return
        if isinstance(Rstar,u.quantity.Quantity):
            self.Rstar=Rstar.to(u.m).value
        else:
            self.Rstar=Rstar

    def compute_density(self):
        """Computes the number density (m^-3) profile of the atmosphere
        in the radiative layers
        """
        self.density=np.power(10., self.logp_opac)/(KBOLTZ*self.t_opac)

    def compute_layer_col_density(self):
        """Computes the column number density (molecules/m^2) per
        radiative layer of the atmosphere.

        There are Nlay-1 radiative layers as they go from the midle of a layer to the next.
        """
        factor=N_A/(self.grav * self.Mgas_rad)
        self.dcol_density_rad = np.diff(self.play)*factor[:]

        if self.Rp is not None: #includes the altitude effect if radius is known
            self.compute_altitudes()
            self.dcol_density_rad*=(1.+self.zlev[1:-1]/self.Rp)**2

    def compute_altitudes(self):
        """Compute altitudes of the level surfaces (zlev) and mid layers (zlay).
        """
        Mgas = np.empty(self.Nlay, dtype=float)
        Mgas[1:-1] = 0.5*(self.Mgas_rad[:-1]+self.Mgas_rad[1:])
        Mgas[0] = self.Mgas_rad[0]
        Mgas[-1] = self.Mgas_rad[-1]
        H = RGP*self.tlay/(self.grav*Mgas)
        dlnP = np.diff(self.logplev)*np.log(10.)
        self.zlev = np.zeros_like(self.logplev)
        if self.Rp is None:
            self.dz = H*dlnP
            self.zlev[:-1] = np.cumsum(self.dz[::-1])[::-1]
        else:
            for i in range(H.size)[::-1]:
                z1 = self.zlev[i+1]
                H1 = H[i]
                dlnp = dlnP[i]
                self.zlev[i] = z1+( (H1 * (self.Rp + z1)**2 * dlnp) \
                    / (self.Rp**2 + H1 * self.Rp * dlnp + H1 * z1 * dlnp) )
        self.zlay = 0.5*(self.zlev[1:]+self.zlev[:-1])
        self.zlay[-1] = 0.
        self.zlay[0] = self.zlev[0]
        ## assumes layer centers at the middle of the two levels
        ## which is not completely consistent with play, but should be
        ## a minor error.
        
    def compute_area(self):
        """Computes the area of the annulus covered by each
        radiative layer (from a mid layer to the next) in a transit setup. 
        """
        self.area=PI*(self.Rp+self.zlay[:-1])**2
        self.area[:-1]-=self.area[1:]
        self.area[-1]-=PI*self.Rp**2

    def compute_tangent_path(self):
        """Computes a triangular array of the tangent path length (in m) spent in each
        radiative layer.
        
        self.tangent_path[ilay][jlay] is the length that the ray that is tangent to the ilay 
        radiative layer spends in the jlay>=ilay layer
        (accounting for a factor of 2 due to symmetry)
        """
        if self.Rp is None: raise RuntimeError('Planetary radius should be set')
        self.compute_altitudes()
        self.tangent_path=List()
        # List() is a new numba.typed list to comply with new numba evolution after v0.50
        for ilay in range(self.Nlay-1): #layers counted from the top
            z0square=(self.Rp+self.zlev[ilay+1])**2
            dl=np.sqrt((self.Rp+self.zlay[:ilay+1])**2-z0square)
            dl[:-1]-=dl[1:]
            self.tangent_path.append(2.*dl)

    def __repr__(self):
        """Method to output header
        """
        output="""
    gravity (m/s^2) : {grav}
    Planet Radius(m): {rad}
    Ptop (Pa)       : {ptop}
    Psurf (Pa)      : {psurf}
    Tsurf (K)       : {tsurf}
    composition     :
        {comp}""".format(grav=self.grav, rad=self.Rp, comp=self.gas_mix,
            ptop=self.plev[0], psurf=self.psurf, tsurf=self.tlay[-1])
        return output

    def plot_T_profile(self, ax, invert_p = True, use_altitudes = False,
            xscale=None, yscale=None, **kwarg):
        """Plot the T P profile
        
        Parameters
        ----------
            ax : :class:`pyplot.Axes`
                A pyplot axes instance where to put the plot.
            x/yscale: str, optional
                If 'log' log axes are used.
        """
        if use_altitudes:
            self.compute_altitudes()
            ax.plot(self.tlay,self.zlay,**kwarg)
            ax.set_ylabel('Altitude (m)')
        else:
            ax.plot(self.tlay,self.play,**kwarg)
            if invert_p: ax.invert_yaxis()
            ax.set_ylabel('Pressure (Pa)')
        ax.set_xlabel('Temperature (K)')
        if xscale is not None: ax.set_xscale(xscale)
        if yscale is not None: ax.set_yscale(yscale)


    def write_soundings(self, dirname='.', fmt='%.10e', cp=None, qvap=None, p_dry=None):
        """Writes sounding files that can be used to initiate the mesoscale model
        """
        self.compute_altitudes()
        mgas = np.zeros(self.Nlay)
        mgas[:-1] = self.Mgas_rad
        mgas[-1] = self.Mgas_rad[-1]
        r = RGP/mgas
        rho_lay = self.play/(r*self.tlay)
        if cp is not None:
            cp_array = cp * np.ones(self.Nlay)
        else:
            cp_array =  r / self.rcp
        zeros=np.zeros(self.Nlay)
        teta = self.tlay / self.exner
        if qvap is not None:
            q=qvap
        else:
            q=zeros
        filename=dirname+'/input_sounding'
        np.savetxt(filename, np.transpose([self.zlay[::-1], teta[::-1], q[::-1], zeros ,zeros]),
            fmt=fmt, header=str(self.psurf/100.)+'  '+str(self.tlay[-1])+'    0.0', comments=' ')
        # the last dummy column are q_vap (not used as of 2021, u, and v). We put it to zero.

        filename=dirname+'/input_therm'
        if p_dry is None:
            np.savetxt(filename,
                np.transpose([r[::-1], cp_array[::-1], self.play[::-1], rho_lay[::-1], self.tlay[::-1]]), fmt=fmt)
        else:
            np.savetxt(filename,
                np.transpose([r[::-1], cp_array[::-1], p_dry[::-1], rho_lay[::-1], self.tlay[::-1]]), fmt=fmt)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['exo_k', 'exo_k.atm_evolution', 'exo_k.two_stream', 'exo_k.util']

package_data = \
{'': ['*']}

install_requires = \
['astropy>=4.3.1',
 'h5py>=3.7.0',
 'llvmlite>=0.38.1',
 'matplotlib>=3.5.2',
 'numba>=0.55.2',
 'numpy>=1.21.6',
 'pandas>=1.3.5',
 'scipy>=1.7.3']

setup_kwargs = {
    'name': 'exo-k',
    'version': '1.2.0',
    'description': 'Library to handle radiative opacities from various sources for atmospheric applications',
    'long_description': "# Exo_k\n\nAuthor: Jeremy Leconte (CNRS/LAB/Univ. Bordeaux)\n\n`Exo_k` is a Python 3 based library to handle radiative opacities from various sources for atmospheric applications.\nIt enables you to:\n\n* Interpolate efficiently and easily in correlated-k and cross section tables.\n* Convert easily correlated-k and cross section tables from one format to another\n  (hdf5, LMDZ GCM, Exomol, Nemesis, PetitCode, TauREx, ExoREM, ARCIS, etc.).\n* Adapt precomputed correlated-k tables to your needs by changing:\n\n  * the resolution and quadrature (g) grid,\n  * the pressure/temperature grid.\n* Create tables for a mix of gases using tables for individual gases.\n* Create your own tables from high-resolution spectra (for example from K-spectrum, Helios-K, etc.).\n* Use your data in an integrated radiative transfer framework to simulate planetary atmospheres.\n  \nFor a complete online documentation, checkout:\nhttp://perso.astrophy.u-bordeaux.fr/~jleconte/exo_k-doc/index.html\n\nIn this repository, you'll find a [tutorial jupyter notebook](https://forge.oasu.u-bordeaux.fr/jleconte/exo_k-public/-/blob/public/tutorial-exo_k.ipynb) that will show you how to do all that\nwith concrete examples that you can run on your own machine. Many important concepts and options are\npresented along the way.\n\nEnjoy!\n\nJ. Leconte\n\n# Acknowledgements\n\nIf you use this library in your research, please acknowledge it by citing\n[Leconte (2021)](https://ui.adsabs.harvard.edu/abs/2021A%26A...645A..20L/abstract):\n\n  * Spectral binning of precomputed correlated-k coefficients. **Astronomy and Astrophysics** 645. Leconte, J. 2021. doi:10.1051/0004-6361/202039040\n\nThis project has received funding from the European Research Council (ERC)\nunder the European Union's Horizon 2020 research and innovation programme\n(grant agreement n° 679030/WHIPLASH).\n\nThe framework for this documentation has been developped by Aurelien Falco using Sphinx.\nThe Framework for automatic testing has been developped by Alexandre Mechineau. \n\n# last release (see past releases below)\n\nv1.2.0 (July 2022): The model for atmospheric evolution is finally stable and documented.\nThe atm module has also seen several note worthy additions: surface albedo, oceans. \nWe also added a framework for an automatic test suite. In particular, we can test several python versions. Additional tests should rapidly come along.\nRosseland and Planck mean opacities can now be computed from radiative tables. \n\n# Installation\n\nExo_k can be installed using pip (without cloning the repository;\ndependencies should be downloaded automatically):\n```\npip install exo_k\n```\nOr by running the [setup.py](https://forge.oasu.u-bordeaux.fr/jleconte/exo_k-public/-/blob/public/setup.py) script in the cloned repository:\n```\npython setup.py install\n```\n\n# Usage\n\nTo learn how to use `exo_k`, you can follow the [tutorial jupyter notebook](https://forge.oasu.u-bordeaux.fr/jleconte/exo_k-public/-/blob/public/tutorial-exo_k.ipynb).\n\nHave fun!\n\n# Links\n\n* Project homepage: http://perso.astrophy.u-bordeaux.fr/~jleconte/\n* Code repository: https://forge.oasu.u-bordeaux.fr/jleconte/exo_k-public\n* Documentation: http://perso.astrophy.u-bordeaux.fr/~jleconte/exo_k-doc/index.html\n* Contact: jeremy.leconte at u-bordeaux.fr\n\n\n# past releases\n\nv1.1.0 (August 2021): New scheme for the computation of atmospheric emission/transmission\nto ensure an improved numerical accuracy. The variable names to instantiate atm objects have\nchanged accordingly (see tutorial). \n\nv1.0.2 (June 2021): Adds a few missing dependencies. Enables computation of thermal\nemission spectra with scattering through the two-stream method (full documentation pending). \nEnables creating Xtables for a mix of gases (CIA can be added as well). Solves some issues\nwith the 2018 Hitran CIA format.\n\nv1.0.1 (Jan 2021): Solves a binary/string conversion issue introduced by version 3 of h5py.\nEnables linear interpolation in pressure (default is log). Enables creation of\nempty tables to be filled later and extension of the spectral range of existing tables. \n\nv1.0.0 (Dec 2020): Finally our first official version. Creation of a\n'examples' notebook with fully worked out use cases for the `Exo_k`. \n\nv0.0.5 (Oct 2020): Ensures compatibility with latest Exomol correlated-k and cross-section tables.",
    'author': 'Jeremy Leconte',
    'author_email': 'jeremy.leconte@u-bordeaux.fr',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://forge.oasu.u-bordeaux.fr/jleconte/exo_k-public',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<3.10',
}


setup(**setup_kwargs)



"""
@author: jeremy leconte

A class with some basic functions for all objects with a spectral dimension
"""
import numpy as np

class Spectral_object(object):
    """A class with some basic functions for all objects with a spectral dimension
    """

    @property
    def wls(self):
        """Returns the wavelength array for the bin centers (in micron)
        """
        if self.wns is not None:
            return 10000./self.wns
        else:
            return None

    @property
    def wledges(self):
        """Returns the wavelength array for the bin edges (in micron)
        """
        if self.wnedges is not None:
            return 10000./self.wnedges
        else:
            return None

    @property
    def wnrange(self):
        """Returns the limits of the wavenumber range.

        First tries with wnedges (for Ktables) and then wns (Xtables).
        """
        if self.wnedges is not None:
            return self.wnedges[[0,-1]]
        elif self.wns is not None:
            return self.wns[[0,-1]]
        else:
            return None

    @property
    def wlrange(self):
        """Returns the limits of the wavelength range
        """
        if (self.wnedges is None) and (self.wns is None):
            return None
        else:
            return np.sort(10000./self.wnrange)

    def select_spectral_range(self, wn_range=None, wl_range=None):
        """Select spectral range, without restricting the data. Should use either wn_range OR wl_range, not both. To be selected, the whole bin should be inside the range.

        Parameters
        ----------
            wn_range: array
                Wavenumber range in cm^-1.
            wl_range: array
                Wavelength range in micron.

        Returns
        -------
            tuple:
                iw_min, iw_max the boundary indices of the spectral range
        """
        if (wn_range is None) and (wl_range is None): return None, None
        if wl_range is not None:
            if wn_range is not None:
                raise RuntimeError('Should provide either wn_range or wl_range, not both!')
            _wn_range=np.sort(10000./np.array(wl_range))
        else:
            _wn_range=np.sort(np.array(wn_range))
        iw_min=np.searchsorted(self.wnedges, _wn_range[0], side='left')
        iw_max=np.searchsorted(self.wnedges, _wn_range[1], side='right')
        iw_max-=1
        if iw_max <= iw_min:
            raise RuntimeError(f"Spectral range {wn_range} does not contain any point.")
        self.wnedges=self.wnedges[iw_min:iw_max+1]
        self.wns=self.wns[iw_min:iw_max]
        self.Nw=self.wns.size
        return iw_min, iw_max
